--[[
This cutscene manager has a simple logic. 

Every action relies on two flags : 

a completed flag = set true when the action is completed
a wait flag = that, if set to true, break the execution loop until that very action is completed.

The wait flag is passed always as last argument but you can code it the way you like.

The wait flag halts the execution of subsequent actions becoming your power to time the scene. 

For example if we have this 5 actions:

animate.wait = false
say.wait = false
moveto.wait = true --breaks here until this one is completed
hide.wait = false
animate.wait = false

the first 3 actions will be updated each tick until completed, once we reach the 3rd action, with a wait flag set to true, we break the loop so the last two ones won't be updated until "moveto" is completed.

The result is that the character is animated and moved where it have to, while saying what it has to say. 
Once it reaches its destination the "moveto" action will mark itself as completed and as such the flow will continue. 

At that point the character will be hidden and a new animation will be played.

This system lets you have multiple actions updating and a "break" point for the scene to proceed after that.


--]]


action = {
  completed = false,
  wait = false
  }

action.__index = action


--[[executeactions must be called from the scene's update and will simply cycle every actions in the passed list until all are completed or a wait flag is found.
I have it as global since is called from a lot of entities.

Basically every entity can have his own actiontable and pass it this function giving the power to execute paraller actions indipendently.

Anyway the simplest way is to just have your "stage" or scene manager with a main action-list and direct everything from there, like I did in this example.
--]]

function executeactions(dt, actiontable,printstatus)

    local allactionscompleted = true
    for k,v in ipairs(actiontable) do

        if v.completed == false then
            v:update(dt)
            allactionscompleted = false
            --break if the action has a wait flag
            if v.wait == true then break end
        end
    end
          
  return allactionscompleted
end

--needed if some actions need to draw stuff.
function drawactions(actiontable)
    for k,v in ipairs(actiontable) do
        if v.completed == false then 
          v:draw()
          if v.wait == true then break end
        end
    end 
end


-- called when adding a new action to the table
function action:new(...)
  local newaction = {}
  setmetatable(newaction, action)
  newaction.args = {...}
  newaction.argn = select("#", ...)
  newaction:init()
  return newaction
end

--this function is used only when passing optional parameters for a reason:
--I have the wait flag always as last argument so to don't be confused I use this for opional params.
function action:getarg(argnumber)
    --last arg is always wait
    if argnumber < self.argn then
        return self.args[argnumber]
    end
    return nil
end

--the init function creates the action itself based on the argument passed setting all its values
function action:init(  )

    --type is the action name
    self.type = self.args[1]
    
    if self.type == "moveto" then
    self.actor = self.args[2] --who to move
    self.target = self.args[3] --another actor or a a table with coords.
    
    --optional speed
    self.movementspeed = self:getarg(4) or self.actor.movementspeed
    
    --moveto auto animate based on direction, this flag is to lock the animation instead.
    self.locked = self:getarg(5) or false
    
    self.wait = self.args[self.argn] or false  
    
    --the update function points to a local function that handles all the stuff to do.
    self.update = self.moveto 
    
        
    elseif self.type == "say" then
    self.firstinit = false
    self.actor = self.args[3]
    self.text= self.args[2]
    self.update = self.say
    self.wait = self.args[self.argn] or false 


   elseif self.type == "animate" then
    self.firstinit = false
    self.actor = self.args[2] 
    self.update = self.animate
    self.animation = self.args[3]
    self.loop = self:getarg(4) or false
    self.wait = self.args[self.argn] or false  --args[5] or false
    

   elseif self.type == "waitasec" then
    self.timer = self.args[2]
    self.wait = true
    self.update = self.waitasec
    
 
    elseif self.type == "hide" then
    self.actor = self.args[2] 
    self.hiding = type(self.args[3]) == "boolean" and self.args[3] or nil
    self.time = type(self.args[3]) == "number" and self.args[3] or nil
    self.wait = type(self.args[3]) == "number" and self.args[self.argn] or false  
    self.update = self.hide 
   
    
    elseif self.type == "fadeout" then
    self.step = self.args[2]
    self.fadeouttime = self:getarg(3) or 1 --default 1 second
    self.wait = self.args[self.argn] or false  --args[self.argn] or false
    self.color = {0,0,0}
    self.update = self.fadeout
    self.draw = self.fadeoutdraw
    

    elseif self.type == "teleport" then
    self.actor = self.args[2] 
    self.target = type(self.args[3]) == "string" and stage:getanchor(self.args[3]) or self.args[3]
    self.update = self.teleport
    
    
    elseif self.type == "playsfx" then
    self.track = self.args[2] 
    self.loop = self.args[3] or false
    self.update = self.playsfx 
    
   
    else
        self.completed = true

  end
end


function action:update(dt)
end
function action:draw()
end



------------------------------------------------------------------------------
--[[Every action is  coded in the section below. this is where the things get done
    just remember to set the action.update = self.functiondoingtuff in the init above.
    like action.update = self.moveto
    so when the action will be updated this function will be executed.

--]]
-----------------------MOVETO-----------------------
function action:moveto(dt) 
  
----------------------------------    

      if not self.firstinit then

        self.x = self.target.x
        self.y = self.target.y

        
        self.firstinit = true
        
        self.actor.animation = self.actor
        
        local distancex = self.actor.animation.x - self.x
        local distancey = self.actor.animation.y - self.y

        self.movementspeed = self.movementspeed or 50
    
        self.xdir = distancex < 0 and self.movementspeed or -self.movementspeed
        self.ydir = distancey < 0 and self.movementspeed or -self.movementspeed
        
        --set direction based on greater distance
        self.animationstate = ""
        if math.abs(distancex) > math.abs(distancey) then
            self.animationstate = self.xdir > 0 and "right" or "left"
        else
            self.animationstate = self.ydir > 0  and "down" or "up" 
        end

    end

    local px,py = self.actor.animation:getPosition()

    local cx = math.floor(px - self.x)
    local cy = math.floor(py - self.y)
    
    
    local completedx = true 
    if self.xdir > 0 and px < self.x then completedx = false end
    if self.xdir < 0 and px > self.x then completedx = false end
    if not completedx then 
        px = px + self.xdir * dt
    end
    
    local completedy = true 
    if self.ydir > 0 and py < self.y then completedy = false end
    if self.ydir < 0 and py > self.y then completedy = false end
    if not completedy then
        py = py + self.ydir * dt
    end

    local updir = math.abs(cx) - math.abs(cy)
    
    self.finaldirection = self.animationstate

    self.actor.animation:setPosition( px, py ) 

    if completedx and completedy then
        self.actor.animation:setPosition( self.x , self.y ) 

        if self.addtoworld then
            local bx, by, bw, bh = self.actor.animation:getHotSpotGeometry("boundbox")
            world:add(self.actor,  self.x + bx,  self.y + by, bw, bh )
        end
        

        local idle
        local state = self.finaldirection
        if state == "left" or state == "runleft" then 
            idle = "idleleft" 
        else
            idle = "idleright" 
        end
        idle = state == "up" and "idleup" or idle
        if not self.locked then
            self.actor.animation:setState(idle, true )
        end
        
        self.completed =  true
    else
        --if animation is locked it won't be changed, even at the end
        if not self.locked then
            self.actor.animation:setState(self.animationstate,true)
        end
    end
  

end


-----------------------SAY-----------------------
function action:say(dt)
    if self.firstinit == false then
        self.draw = self.saydraw
        local x, y
        
        self.x, self.y = self.actor:getHotSpotCoords("overhead")

        self.firstinit = true
        self.draw = self.saydraw

        return
    end

    
    --for simplicity in this exmple let's just wait for a keypress and then "consume it" but you wont do this...
    if scene.actionkey then
        scene.actionkey = false
        self.completed = true
    end
 
end

function action:saydraw()
    local font = love.graphics.getFont( )
    local width = font:getWidth( self.text )
    local height = font:getHeight( self.text )

    --self.x and self.y comes from an overhead hotspot in the animation
    --so i put the text up there just centered.
    love.graphics.print(self.text, self.x - width / 2 ,self.y - height )
end




-----------------------ANIMATE-----------------------
function action:animate (dt)
        if self.firstinit == false then
            self.actor:setState(self.animation, self.loop)
            self.firstinit = true
            self.name = "action"
            self.actor:onStateEnd( self.animation , self , self.animateend )
          return
        end
end
function action:animateend()
        self.completed = true
end





-----------------------WAIT-----------------------

function action:waitasec(dt)
    self.timer = self.timer - dt
    if self.timer <= 0 then
        self.completed = true
    end  
end



-----------------------HASTE-----------------------
function action:haste(dt)
    self.actor:setSpeedMultiplier(self.animationspeed)
    self.completed = true
end


-----------------------HIDE-----------------------

function action:hide (dt)
    if not self.firstinit then
        self.firstinit = true
        
        if not self.time then
            if self.actor.alpha < 1 and not self.hiding then
                self.actor.alpha  = 1
            else
                self.actor.alpha  = 0
            end
            
            self.completed = true
            return
        else
            self.step = 1 / self.time 
            self.actor.alpha = self.time > 0 and 0 or 1
            self.time = math.abs(self.time)
        end
    end
    
                
    self.time = self.time - dt
    self.actor.alpha = self.actor.alpha + self.step * dt
    
    if self.time <= 0 then
        self.actor.alpha = self.step < 0 and 0 or 1
        self.completed = true
    end
end


-----------------------FADEOUT-----------------------

function action:fadeout(dt)
    if not self.firstinit then
        self.firstinit = true
        
        self.alpha =  self.step == 1 and 0 or 1
        self.fadeamount = 1 / self.fadeouttime * self.step
    
        scene.curtain = self 
        return
    end



    self.alpha = self.alpha  +  self.fadeamount * dt
    
    if self.alpha >= 1 and self.step == 1 then
        self:fadeoutend("close")
    elseif self.alpha <= 0 and self.step == -1 then
        self:fadeoutend("open")
    end
        
    
end

function action:fadeoutdraw()

    local color = {love.graphics.getColor()}
    love.graphics.setColor( self.color[1], self.color[2], self.color[3] , self.alpha )
    love.graphics.rectangle("fill",0,0,love.graphics.getWidth(),love.graphics.getHeight() )
    love.graphics.setColor(color)
    
end
function action:fadeoutend(state)
    if state == "open" then
        scene.curtain = nil
    elseif state == "close" then
        
    end


    self.completed = true
end




-----------------------TELEPORT-----------------------

function action:teleport(dt)
    
    self.actor:setPosition(self.target.x,self.target.y)

    self.completed = true
end



-----------------------PLAYSFX-----------------------
--I use an audio manager but.. for this example let's jut create a source and  play
function action:playsfx(dt)
        self.source =  love.audio.newSource(self.track, "static")

        self.source:play()
        self.completed=true

end
