--[[
The MIT License (MIT)

Copyright (c) 2023 The Starknight - Denis

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

]]

--[[ aseprite is the library I wrote to load and use aseprite files.
Is fast, it works, and it is easy to implement using the aseprite specs from the site.
I really love how it came out, anyway some data is skipped at loading time but you can uncomment those lines to activate it.
Other parts are not implemented since I didn't needed those. (like exporting tilemaps from aseprite.)
One day maybe I will make a generic one, or feel free to do it yourself.

-denis

--]]

aseprite = {
structs = {}
}

aseprite.__index = aseprite


-- https://www.lua.org/manual/5.3/manual.html#6.4.2
local byteformat = {
 BYTE  = { packstring =  "B",  size= 1 },
 WORD  = { packstring =  "H",  size= 2 },
 SHORT = { packstring =  "h",  size= 2 },
 DWORD = { packstring = "I4",  size= 4 },
 LONG  = { packstring = "i4",  size= 4 },
 FIXED = { packstring = "i4",  size= 4 }
}


local BYTE  =  "BYTE"
local WORD  =  "WORD"
local SHORT =  "SHORT" 
local DWORD =  "DWORD"
local LONG  =  "LONG" 
local FIXED =  "FIXED" 
local STRING = "WORD" -- the string length


local header_struct = {
DWORD , --      File size
WORD , --        Magic number (0xA5E0)
WORD , --       Frames
WORD , --        Width in pixels
WORD , --       Height in pixels
WORD , --       Color depth (bits per pixel)
      --        32 bpp = RGBA
      --        16 bpp = Grayscale
      --        8 bpp = Indexed
DWORD , --       Flags:
        --      1 = Layer opacity has valid value
WORD , --        Speed (milliseconds between frame, like in FLC files)
       --     DEPRECATED: You should use the frame duration field
       --     from each frame header
DWORD , --      Set be 0
DWORD , --      Set be 0
BYTE  , --      Palette entry (index) which represent transparent color
        --    in all non-background layers (only for Indexed sprites).
BYTE , --    Ignore these bytes
BYTE ,
BYTE ,
WORD  , --      Number of colors (0 means 256 for old sprites)
BYTE , --       Pixel width (pixel ratio is "pixel width/pixel height").
        --    If this or pixel height field is zero, pixel ratio is 1:1
BYTE   , --     Pixel height
SHORT , --      X position of the grid
SHORT , --      Y position of the grid
WORD, --     Grid width (zero if there is no grid, grid size
         --   is 16x16 on Aseprite by default)
WORD  --       Grid height (zero if there is no grid)
}

local frame_struct = {
DWORD , --       Bytes in this frame
WORD  , --       Magic number (always 0xF1FA)
WORD  , --       Old field which specifies the number of "chunks"
        --       in this frame. If this value is 0xFFFF, we might
        --       have more chunks to read in this frame
        --       (so we have to use the new field)
WORD  , --       Frame duration (in milliseconds)
BYTE  , --[2]    For future (set to zero)
BYTE  ,
DWORD  --        New field which specifies the number of "chunks"
       --        in this frame (if this is 0, use the old field)

}

local chunk_struct = {
DWORD ,--       Chunk size
WORD  --       Chunk type
}

local rgb_struct = {
BYTE ,--    Red (0-255)
BYTE ,--    Green (0-255)
BYTE  --    Blue (0-255)
}

local rgba_struct = {
BYTE ,--    Red (0-255)
BYTE ,--    Green (0-255)
BYTE ,--    Blue (0-255)
BYTE  --    Alpha (0-255)
}

local layer_struct = {

WORD ,--        Flags:
      --        1 = Visible
      --        2 = Editable
      --        4 = Lock movement
      --        8 = Background
      --        16 = Prefer linked cels
      --        32 = The layer group should be displayed collapsed
      --        64 = The layer is a reference layer
WORD ,--        Layer type
      --        0 = Normal (image) layer
      --        1 = Group
      --        2 = Tilemap
WORD ,--        Layer child level (see NOTE.1)
WORD ,--        Default layer width in pixels (ignored)
WORD ,--        Default layer height in pixels (ignored)
WORD ,--        Blend mode (always 0 for layer set)
      --        Normal         = 0
      --        Multiply       = 1
      --        Screen         = 2
      --        Overlay        = 3
      --        Darken         = 4
      --        Lighten        = 5
      --        Color Dodge    = 6
      --        Color Burn     = 7
      --        Hard Light     = 8
      --        Soft Light     = 9
      --        Difference     = 10
      --        Exclusion      = 11
      --        Hue            = 12
      --        Saturation     = 13
      --        Color          = 14
      --        Luminosity     = 15
      --        Addition       = 16
      --        Subtract       = 17
      --        Divide         = 18
BYTE ,--        Opacity
      --        Note: valid only if file header flags field has bit 1 set
BYTE ,--[3]     For future (set to zero)
BYTE ,
BYTE ,
STRING
}
--[[STRING      Layer name
+ If layer type = 2
  DWORD     Tileset index

}
]]


local cell_struct = {
WORD  ,--        Layer index (see NOTE.2)
SHORT ,--        X position
SHORT ,--        Y position
BYTE  ,--        Opacity level
WORD  ,--        Cel Type
       --        0 - Raw Image Data (unused, compressed image is preferred)
       --        1 - Linked Cel
       --        2 - Compressed Image
       --        3 - Compressed Tilemap
SHORT ,--        Z-Index (see NOTE.5)
       --        0 = default layer ordering
       --        +N = show this cel N layers later
       --        -N = show this cel N layers back
BYTE  ,-- 5]     For future (set to zero)
BYTE  ,
BYTE  ,
BYTE  ,
BYTE  
}

local colorprofile_struct = {
WORD  ,--        Type
       --       0 - no color profile (as in old .aseprite files)
       --       1 - use sRGB
       --       2 - use the embedded ICC profile
WORD  ,--       Flags
       --       1 - use special fixed gamma
FIXED ,--       Fixed gamma (1.0 = linear)
       --     Note: The gamma in sRGB is 2.2 in overall but it doesn't use
       --     this fixed gamma, because sRGB uses different gamma sections
       --     (linear and non-linear). If sRGB is specified with a fixed
       --     gamma = 1.0, it means that this is Linear sRGB.
BYTE  ,--  [8]     Reserved (set to zero)
BYTE  ,
BYTE  ,
BYTE  ,
BYTE  ,
BYTE  ,
BYTE  ,
BYTE  
}
--[[
+ If type = ICC:
  DWORD     ICC profile data length
  BYTE[]    ICC profile data. More info: http://www.color.org/ICC1V42.pdf
]]


--[[
WORD        Number of tags
BYTE[8]     For future (set to zero)
]]


local tag_struct = {
WORD ,--      From frame
WORD ,--      To frame
BYTE ,--      Loop animation direction
      --      0 = Forward
      --      1 = Reverse
      --      2 = Ping-pong
      --      3 = Ping-pong Reverse
WORD ,--      Repeat N times. Play this animation section N times:
      --      0 = Doesn't specify (plays infinite in UI, once on export,
      --        for ping-pong it plays once in each direction)
      --      1 = Plays once (for ping-pong, it plays just in one direction)
      --      2 = Plays twice (for ping-pong, it plays once in one direction,
      --        and once in reverse)
      --      n = Plays N times
BYTE ,BYTE ,BYTE ,BYTE ,BYTE ,BYTE ,--[6]   For future (set to zero)
BYTE ,BYTE ,BYTE ,--[3]   RGB values of the tag color
      --      Deprecated, used only for backward compatibility with Aseprite v1.2.x
      --      The color of the tag is the one in the user data field following
      --      the tags chunk
BYTE ,--      Extra byte (zero)
STRING --    Tag name
}

local palette_struct = {
DWORD ,--       New palette size (total number of entries)
DWORD ,--       First color index to change
DWORD ,--       Last color index to change
BYTE , BYTE , BYTE , BYTE ,--8]     For future (set to zero)
BYTE , BYTE , BYTE , BYTE 
}

--+ For each palette entry in [from,to] range (to-from+1 entries)
local paletteentry_struct = {
WORD ,--      Entry flags:
      --    1 = Has name
BYTE ,--      Red (0-255)
BYTE ,--      Green (0-255)
BYTE ,--      Blue (0-255)
BYTE --       Alpha (0-255)
}
--[[
+ If has name bit in entry flags
STRING  Color name
]]


--I use only text for userdata
local userdata_struct = {
DWORD --       Flags
       --       1 = Has text
       --       2 = Has color
       --       4 = Has properties
}
--[[
+ If flags have bit 1
  STRING    Text
+ If flags have bit 2
  BYTE      Color Red (0-255)
  BYTE      Color Green (0-255)
  BYTE      Color Blue (0-255)
  BYTE      Color Alpha (0-255)

+ If flags have bit 4
  DWORD     Size in bytes of all properties maps stored in this chunk
            The size includes the this field and the number of property maps
            (so it will be a value greater or equal to 8 bytes).
  DWORD     Number of properties maps
  + For each properties map:
    DWORD     Properties maps key
              == 0 means user properties
              != 0 means an extension Entry ID (see External Files Chunk))
    DWORD     Number of properties
    + For each property:
      STRING    Name
      WORD      Type
      + If type==0x0001 (bool)
        BYTE    == 0 means FALSE
                != 0 means TRUE
      + If type==0x0002 (int8)
        BYTE
      + If type==0x0003 (uint8)
        BYTE
      + If type==0x0004 (int16)
        SHORT
      + If type==0x0005 (uint16)
        WORD
      + If type==0x0006 (int32)
        LONG
      + If type==0x0007 (uint32)
        DWORD
      + If type==0x0008 (int64)
        LONG64
      + If type==0x0009 (uint64)
        QWORD
      + If type==0x000A
        FIXED
      + If type==0x000B
        FLOAT
      + If type==0x000C
        DOUBLE
      + If type==0x000D
        STRING
      + If type==0x000E
        POINT
      + If type==0x000F
        SIZE
      + If type==0x0010
        RECT
      + If type==0x0011 (vector)
        DWORD     Number of elements
        WORD      Element's type.
        + If Element's type == 0 (all elements are not of the same type)
          For each element:
            WORD      Element's type
            BYTE[]    Element's value. Structure depends on the
                      element's type
        + Else (all elements are of the same type)
          For each element:
            BYTE[]    Element's value. Structure depends on the
                      element's type
      + If type==0x0012 (nested properties map)
        DWORD     Number of properties
        BYTE[]    Nested properties data
                  Structure is the same as indicated in this loop
      + If type==0x0013
        UUID
]]



local layerflags = {
visible = 1,
editable = 2, 
lock_movement = 4, 
background = 8, 
prefer_linked = 16, 
display_group_collapsed = 32, 
reference_layer = 64, 
}

              
local function getstruct( structtable )
    --cache structs for speed
    if aseprite.structs[structtable] then
        return aseprite.structs[structtable].size, aseprite.structs[structtable].packstring
    else
        local size = 0
        local packstring =""
        for _ ,v  in ipairs(structtable) do
            --size = size + byteformat[v].size
            packstring = packstring .. byteformat[v].packstring
        end
        size = love.data.getPackedSize(packstring)
        aseprite.structs[structtable]= {size = size, packstring = packstring }
        return size, packstring
    end
end

local function readunpacked( file, type )
    local val = file:read( byteformat[type].size )
    val = love.data.unpack( byteformat[type].packstring, val )
    return val
end

local function fileskip( file, bytes )
    local pos = file:tell()
    pos = pos + bytes
    file:seek(pos)
end

local function readheader(file, datatable)
	--local datatable = {}

    local size, packstring = getstruct(header_struct)
    local buffer =  file:read(size)

	datatable.file_size,
    datatable.magic_number,
    datatable.frames_number,
    datatable.width,
    datatable.height,
    _ ,
    datatable.flags = love.data.unpack(packstring, buffer)
    	
	if datatable.magic_number ~= 0xA5E0 then
		error("Wrong aseprite magic number")
	end

    --skip next 84
    size = 84 * byteformat.BYTE.size

    fileskip(file, size )
    
	--return datatable
end

local function readframe(file )
	local datatable = {}

    local size, packstring = getstruct(frame_struct)
    local buffer =  file:read(size)

	datatable.frame_size,
    datatable.magic_number,
    _ ,
    datatable.frame_duration,
    _ ,
    _ ,
    datatable.chunks_number = love.data.unpack(packstring, buffer)


	if datatable.magic_number ~= 0xF1FA then
		error("Wrong frame magic number")
	end

	return datatable
end

local function readoldpalette(file)
    --just skip this
--[[
WORD        Number of packets
+ For each packet
  BYTE      Number of palette entries to skip from the last packet (start from 0)
  BYTE      Number of colors in the packet (0 means 256)
  + For each color in the packet
    BYTE    Red (0-255)
    BYTE    Green (0-255)
    BYTE    Blue (0-255)
]]--

    local size = 0
    local numpackets =  readunpacked( file,WORD)
    for p = 1, numpackets do
        _ =  file:read(byteformat.BYTE.size )
        local colors =  readunpacked( file,BYTE) 
        if colors == 0 then colors = 256 end
        size = getstruct(rgb_struct)
        size = size * colors
        fileskip(file, size )
    end
    
	return 
end

local function readlayer(file, asetable)
	local datatable = {}

    local size, packstring = getstruct(layer_struct)
    local buffer =  file:read(size)
    
    local namelength
	datatable.flags,
    datatable.type,
    _ ,
    _ ,
    _ ,
    datatable.blendmode,
    datatable.opacity,
    _ , _ , _ , 
    namelength = love.data.unpack(packstring, buffer)
    
    local on = false
    for k,v in pairs(layerflags)do
        on = false
        if bit.band(datatable.flags,v) ~= 0 then
            on = true        
        end
        datatable[k] = on
    end
    
    --sets index to check if there is only one drawable layer or more, to save time later.
    if datatable.visible == true then
        if asetable.singlelayer == nil then
            --not set, we are the first layer
            asetable.singlelayer = true
        elseif asetable.singlelayer == true then
            --already set, we are not alone, set to false
            asetable.singlelayer = false
        end
    end

    datatable.name = file:read(namelength)
    table.insert(asetable.layers, datatable)
	--return datatable
end

local function readcell(file, datatable )
	--local datatable = {}

    local size, packstring = getstruct(cell_struct)
    local buffer =  file:read(size)

	datatable.layer_index,
    datatable.x,
    datatable.y,
    datatable.opacity,
    datatable.type,
    _,
    _ , _ , _ , _ , _  = love.data.unpack(packstring, buffer)
    
    
    --change to lua format
    datatable.layer_index = datatable.layer_index  + 1
    
    --skip all other types since I only use type 2
    if datatable.type == 2 then
		datatable.width = readunpacked( file,WORD) 
		datatable.height = readunpacked( file,WORD) 
		datatable.image = file:read(datatable.size - 26) --chunk size  -  headers
    else
        error("Wrong aseprite cell type :" .. datatable.type)
	end
    
	--return datatable
end

local function readcolorprofile(file, datatable)
    
    local size, packstring = getstruct(colorprofile_struct)
    local buffer =  file:read(size)
    
    local namelength
	datatable.type,
    datatable.flags,
    datatable.gamma,
     _ , _ , _ , _ ,
     _ , _ , _ , _  = love.data.unpack(packstring, buffer)
    
end

local function readtags(file, asetable)
    --[[
WORD        Number of tags
BYTE[8]     For future (set to zero)
]]
    local number = readunpacked( file, WORD) 
    fileskip(file, 8 )
    
    for t = 1, number do
        local datatable = {}
        local size, packstring = getstruct(tag_struct)
        local buffer =  file:read(size)
        
        local namelength
        datatable.from,
        datatable.to,
        datatable.loop,
        datatable.repeat_times ,
         _ , _ , _ , _ , _ , _ ,
         _ , _ , _ ,
         _ ,
        namelength = love.data.unpack(packstring, buffer)

        datatable.name = file:read(namelength)
        
        asetable.tags[datatable.name] = datatable
    end
end

local function readpalette(file, datatable )
    --skip chunk entirely the fast way
    local size = getstruct(chunk_struct)
    size = datatable.size - size --remove from chunk size the chunk header.size
    fileskip(file, size )
end


local function readuserdata(file, userdatachunk)
--[[
DWORD       Flags
              1 = Has text
              2 = Has color
              4 = Has properties
]]
    local datatable = {}
    local flags =  readunpacked( file,DWORD) 
    if flags ==  1 or flags == 3 then
        local textlength = readunpacked( file,WORD) 
        datatable.text = file:read(textlength)
    end
    
    --sometimes there are colors too, so need to skip it
    if flags ==  2 or flags == 3 then
        local size = getstruct(rgba_struct)
        fileskip(file, size )
    end
    
    if userdatachunk ~= nil then
        userdatachunk.userdata = datatable
    end
    
end

local function readchunk(file, asetable, userdatachunk)
	local datatable = {}

    local size, packstring = getstruct(chunk_struct)
    local buffer =  file:read(size)
    
	datatable.size,
    datatable.chunktype = love.data.unpack(packstring, buffer)
	if datatable.chunktype == 0x2004 then
        readlayer(file, asetable)
        
    elseif datatable.chunktype == 0x2005 then
        readcell(file, datatable ) 
        return datatable

    elseif datatable.chunktype == 0x2018 then
        readtags(file, asetable)
        
    elseif datatable.chunktype == 0x2020 then
        readuserdata(file, userdatachunk )  
        
        
--	elseif datatable.chunktype == 0x0004 then
--        readoldpalette(file)
        
--    elseif datatable.chunktype == 0x2019 then
--        readpalette(file, datatable)
        
--    elseif datatable.chunktype == 0x2007 then
--        readcolorprofile(file, datatable)
        
    else
        --skip chunk
        size = datatable.size - size --remove from chunk size the chunk header.size
        fileskip(file, size )
    end

end


function aseprite.load(asefile, framenumber)
    
    local file = love.filesystem.newFile(asefile)
    
    if file == nil then
        error("File not found :" .. asefile)
    end
    
    -- error if file is not found
    if not file:open("r") then
        error("Can't open :" .. asefile)
    end
    
    print(asefile)
    file:open("r")

	local asetable = {
    frames = {}, --ipairs
    layers = {}, --ipairs
    tags =   {},  --pairs
    singlelayer = nil 
    }
    
	readheader(file, asetable )

    
    local startframe = 1
    
	for f = startframe , asetable.frames_number do
		asetable.frames[f] = readframe(file,f)
        asetable.frames[f].data = {}
        local userdatachunk = nil --pass to link userdata to the right chunk as userdata field
        local dataindex = 0
		for c = 1, asetable.frames[f].chunks_number do
            local imagechunk = readchunk( file, asetable, userdatachunk )
            if imagechunk then
                dataindex = dataindex + 1
                asetable.frames[f].data[dataindex] = imagechunk
                --accept userdata only on images... 
                userdatachunk = asetable.frames[f].data[dataindex]
            end
        end
        --if i reach the framenumber break the cycle
        if framenumber and f == framenumber then break end
    end
    
    --properly sets singlelayer flag
    file:close()
	return asetable
end

